# -*- coding: utf-8 -*-
import numpy as np
import logging
import datetime

import oxdls

from .edge_definer import marker_edge_definer


class MetadataMaker():

    def __init__(self, image_name, unstitched, datatype):
        # Set up attributes and variables
        self.boundaries = unstitched.boundaries
        self.mosaic_centre = unstitched.mosaic_centre
        self.pix2edge = unstitched.pix2edge
        self.pixelsize = unstitched.pixel_size
        mosaic_dims = self.__get_mosaic_dims()
        x_position, y_position = self.__get_x_y_position(self.boundaries, mosaic_dims, unstitched.pixel_size)
        physical_mosaic_dims = [dim * unstitched.pixel_size for dim in mosaic_dims]
        date_time = datetime.datetime.fromtimestamp(unstitched.modified_timestamp).isoformat()  # formatted as: "yyyy-mm-ddThh:mm:ss"

        logging.info("Creating OME metadata")
        self.ox = oxdls.OMEXML()
        image = self.ox.image()
        image.set_Name(image_name)
        image.set_ID("Image:0")
        image.set_AcquisitionDate(date_time)

        pixels = image.Pixels
        pixels.set_DimensionOrder("XYZCT")
        pixels.set_ID("Pixels:0")
        pixels.set_PixelType(str(datatype))
        pixels.set_SizeX(mosaic_dims[0])
        pixels.set_SizeY(mosaic_dims[1])
        pixels.set_SizeZ(1)
        pixels.set_SizeC(1)
        pixels.set_SizeT(1)
        pixels.set_PhysicalSizeX(physical_mosaic_dims[0] * 1.e3)
        pixels.set_PhysicalSizeXUnit("nm")
        pixels.set_PhysicalSizeY(physical_mosaic_dims[1] * 1.e3)
        pixels.set_PhysicalSizeYUnit("nm")
        pixels.set_PhysicalSizeZ(1)  # Z doesn't have corresponding data
        pixels.set_PhysicalSizeZUnit("reference frame")
        pixels.set_tiffdata_count(1)
        pixels.set_plane_count(1)

        channel = pixels.channel(0)
        channel.set_ID("Channel:0:0")
        channel.set_Name("C:0")

        tiffdata = pixels.tiffdata(0)
        tiffdata.set_FirstC(0)
        tiffdata.set_FirstZ(0)
        tiffdata.set_FirstT(0)
        tiffdata.set_IFD(0)

        # Add plane/tiffdata
        plane = pixels.plane(0)
        plane.set_TheZ(0)
        plane.set_TheC(0)
        plane.set_TheT(0)
        plane.set_PositionXUnit("nm")
        plane.set_PositionYUnit("nm")
        plane.set_PositionZUnit("reference frame")
        plane.set_PositionX(x_position * 1.e3)
        plane.set_PositionY(y_position * 1.e3)
        plane.set_PositionZ(0)


    def add_markers(self, updated_image_name, markerfile):
        logging.info("Creating ROIs")
        markerlist, marker_numbers = self.__extract_markers(markerfile)

        no_of_markers = len(marker_numbers)

        image = self.ox.image(0)
        image.set_Name(updated_image_name)
        image.set_roiref_count(no_of_markers)
        self.ox.set_roi_count(no_of_markers)

        for count in range(0, no_of_markers, 1):
            start, end = marker_edge_definer(
               markerlist[count],
                self.boundaries,
                self.pix2edge
                )
            image.roiref(count).set_ID(marker_numbers[count])
            roi = self.ox.roi(count)
            roi.set_ID(marker_numbers[count])
            roi.set_Name(f"Marker {marker_numbers[count]}")

            rectangle = roi.Union.Rectangle()
            rectangle.set_ID(f"Shape:{count}:0")
            rectangle.set_Text(f"Area {marker_numbers[count]}")
            rectangle.set_TheZ(0)
            rectangle.set_TheC(0)
            rectangle.set_TheT(0)

            # Colour is set using RGBA to int conversion
            # RGB colours: Red=-16776961, Green=16711935, Blue=65535
            # Calculated the following function from https://docs.openmicroscopy.org/omero/5.5.1/developers/Python.html:
            # def rgba_to_int(red, green, blue, alpha=255):
            #     # Return the color as an Integer in RGBA encoding
            #     r = red << 24
            #     g = green << 16
            #     b = blue << 8
            #     a = alpha
            #     rgba_int = r+g+b+a
            #     if (rgba_int > (2**31-1)):       # convert to signed 32-bit int
            #         rgba_int = rgba_int - 2**32
            #     return rgba_int
            rectangle.set_StrokeColor(-16776961)  # Red
            rectangle.set_StrokeWidth(20)
            rectangle.set_X(start[0])
            rectangle.set_Y(start[1])
            rectangle.set_Width(end[0] - start[0])
            rectangle.set_Height(end[1] - start[1])

    def get(self, encoded=False):
        if self.ox is not None:
            if encoded:
                return self.ox.to_xml().encode()
            else:
                return self.ox
        logging.error("Cannot get metadata that has not yet been created.")
        return self.ox

    def __extract_markers(self, markerfile):
        # returns marker coordinates in pixels
        array = np.genfromtxt(markerfile, delimiter=",")
        marker_coordinates = []
        marker_numbers = []
        for count in range(len(array[:, 0])):
            x, y = array[count, 0:2]
            # x is flipped between image and marker coordinates
            x = ((-x - self.mosaic_centre[0]) / self.pixelsize)
            y = ((y - self.mosaic_centre[1]) / self.pixelsize)
            marker_coordinates.append((int(x), int(y)))
            marker_numbers.append(int(array[count, -1]))
        return marker_coordinates, marker_numbers

    @staticmethod
    def __get_x_y_position(boundaries, mosaic_dims, pixelsize):
        # (minimum position in x & y + half the image size) * pixel size to get physical positions
        x_position = (boundaries[0, 0] + (mosaic_dims[0] / 2)) * pixelsize
        y_position = (boundaries[0, 1] + (mosaic_dims[1] / 2)) * pixelsize
        return x_position, y_position

    def __get_mosaic_dims(self):
        # In pixels
        dim_x = (self.boundaries[1, 0] - self.boundaries[0, 0])
        dim_y = (self.boundaries[1, 1] - self.boundaries[0, 1])
        return [dim_x, dim_y]



import numpy as np
import logging


def _image_value_trimmer(image_stack):
    logging.info("Trimming brightest pixels")
    median_std = np.median(np.std(image_stack, axis=(1, 2)))
    new_max = np.median(image_stack) + 2.5 * median_std
    return np.clip(image_stack, 0, new_max)


def exposure_correct(images, exposure_minmax, brightfield_image_list):
    """
    Cockpit saves some exposure min/max data which they use to
    correct exposure when displaying it. This function uses this
    to correct for the exposure.
    """
    logging.info("Applying exposure correction")
    images_min = images[brightfield_image_list].min()
    images_max = images[brightfield_image_list].max()
    images_range = images_max - images_min
    # make broadcastable with brightfield_image_list (propagates to multiplier)
    exp_min = exposure_minmax[brightfield_image_list, 0, np.newaxis, np.newaxis]
    exp_max = exposure_minmax[brightfield_image_list, 1, np.newaxis, np.newaxis]
    multiplier = images_range / (exp_max - exp_min)
    return (images[brightfield_image_list] - exp_min) * multiplier


def normalise_to_datatype(images, datatype, trim=True):
    # Trim images before correction to avoid any speckles
    # leading to the entire image to be quite dark:
    if trim:
        images = _image_value_trimmer(images)
    else:
        images = images.astype(np.float64)  # Convert to float for rescaling

    logging.info("Re-scaling images to %s", datatype)
    
    # Move minimum value of all corrected images to 0:
    np.subtract(images, images.min(), out=images)
    # Convert values to float and rescale so the maximum
    # is set by datatype:
    # New max should be 1 less than the max allowed by datatype
    # so that the background (max) can be made transparent without losing data
    new_max = (np.iinfo(datatype).max - 1)
    multiplier = new_max / images.max()
    np.multiply(images, multiplier, out=images)
    return images


def cast_to_dtype(image, data_type):
    try:
        dtype = np.dtype(data_type)
        if np.issubdtype(dtype, np.integer) and np.issubdtype(image.dtype, np.floating):
            dtype_info = np.iinfo(dtype)
            # Round min/max to avoid this warning when the issue is just going to be rounded away.
            img_min, img_max = round(image.min()), round(image.max())
            min_clip = dtype_info.min > img_min
            max_clip = dtype_info.max < img_max
            if min_clip or max_clip:
                if min_clip:
                    logging.warning(
                        "Image min %f below %s minimum of %i, values below this will be cut off",
                        img_min, dtype, dtype_info.min)
                if max_clip:
                    logging.warning(
                        "Image max %f above %s maximum of %i, values above this will be cut off",
                        img_max, dtype, dtype_info.max)
                np.clip(image, dtype_info.min, dtype_info.max, out=image)
        np.around(image, decimals=0, out=image)
        return image.astype(dtype)
    except Exception:
        logging.error("Invalid data type given: %s aka %s. Saving with default data type.", data_type, dtype)
    return image



import os
import logging
import unittest
import pathlib
from pathlib import Path
from unittest.mock import patch, MagicMock

import stitch_m
from stitch_m.file_handler import create_user_config, create_Windows_shortcut, _create_lnk_file, _get_desktop_path

class TestSetupFunctions(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Set maximum difference string length to None (infinite)
        cls.maxDiff = None

    # ------------------
    # Test config setup:
   
    def test_setup_config(self):
        with patch('shutil.copyfile', MagicMock()) as mocked_copyfile:
            local_config_file = Path(stitch_m.__file__).resolve().with_name("config.cfg")
            user_config_location = Path(stitch_m.__file__).parent / "test_config_path"
            with patch('stitch_m.file_handler.get_user_config_path', MagicMock(return_value=(user_config_location, []))):
                create_user_config()
            mocked_copyfile.assert_called_once_with(local_config_file, user_config_location)

    @patch('stitch_m.file_handler.get_user_config_path')
    @patch('logging.error')
    def test_setup_config_fail_bad_path(self, mocked_error_log, mocked_get_config):
        mocked_get_config.return_value = (Path(os.path.expanduser("~/.fake_dir/oh_no/thisisbad.cfg")), [])
        create_user_config()
        mocked_error_log.assert_called_once_with("Unable to create user config file due to directory issues", exc_info=True)

    # ------------------
    # Test Windows shortcut setup:

    @patch("os.name")
    @patch('logging.error')
    def test_setup_win_exits_on_linux(self, mocked_error_log, mocked_os_name):
        mocked_os_name.return_value = "posix"
        create_Windows_shortcut()
        mocked_error_log.assert_called_once_with("This command is only valid on Windows installations.")

    @patch.object(pathlib.Path, "exists", MagicMock(return_value=False))
    @patch('logging.error')
    def test_setup_windows_shortcut_function_called(self, mocked_error):
        with patch('stitch_m.file_handler._create_lnk_file', MagicMock()) as mocked_shortcut_creator:
            create_Windows_shortcut()
            if os.name == "nt":
                home_dir = Path(os.environ["HOMEDRIVE"]) / os.environ["HOMEPATH"]
                try:
                    desktop = _get_desktop_path()
                    self.assertTrue(desktop.relative_to(home_dir), "Invalid desktop found")
                except Exception:
                    logging.warning("_get_desktop_path failed")
                    desktop = home_dir / "Desktop"

                self.assertTrue(desktop.is_dir(), "Invalid desktop found")
                mocked_shortcut_creator.assert_called_once_with(desktop / "StitchM.lnk")
            else:
                mocked_error.assert_called_once_with("This command is only valid on Windows installations.")

    @unittest.skipUnless(os.name == "nt", "Only run on Windows")
    def test_setup_windows_shortcut_test_created_(self):
        # Only run this test if on Windows
        test_shortcut_path = Path(".") / "test_shortcut.lnk"
        _create_lnk_file(test_shortcut_path)
        link_created = test_shortcut_path.exists()
        if link_created:
            os.remove(test_shortcut_path)
        self.assertTrue(link_created)



import unittest
from unittest.mock import patch, MagicMock, call
from tempfile import TemporaryDirectory

from pathlib import Path
from io import StringIO

from stitch_m.log_handler import LogHandler
    

class TestLoggerMethods(unittest.TestCase):

    @staticmethod
    def create_logger(log_dir, log_to_file):
        with patch('stitch_m.file_handler.get_user_log_path', MagicMock(return_value=(log_dir))):
            return LogHandler(file_level="info", stream_level="error", log_to_file=log_to_file, backup_count=1)

    def test_logger_without_file(self):
        with TemporaryDirectory(prefix="logs", dir=".") as log_dir:
            log_dir = Path(log_dir)
            log_handler = TestLoggerMethods.create_logger(log_dir, False)
            log_files = list(log_dir.glob("*"))
            log_handler.close()
            self.assertFalse(log_files)           

    @patch('sys.stdout', new_callable=StringIO)
    def test_loggers_created(self, mocked_stdout):
        with TemporaryDirectory(prefix="logs_", dir=Path.cwd()) as log_dir:
            log_dir = Path(log_dir).absolute()
            log_handler = TestLoggerMethods.create_logger(log_dir, True)
            logger = log_handler.get_logger()

            debug_line = "This is debug info"
            info_line = "This is info"
            error_line = "This is an error"
            logger.debug(debug_line)
            logger.info(info_line)
            logger.error(error_line)

            log_handler.close()

            log_files = tuple(log_dir.glob("*"))
            self.assertEqual(len(log_files), 1)
            for log in log_files:
                with log.open("r") as f:
                    lines = f.read()
                    self.assertNotIn(debug_line, lines)
                    self.assertIn(info_line, lines)
                    self.assertIn(error_line, lines)
        stdout = mocked_stdout.getvalue()
        self.assertNotIn(debug_line, stdout)
        self.assertNotIn(info_line, stdout)
        self.assertIn(error_line, stdout)

    def test_file_path_fails(self):
        # Shouldn't throw an error:
        TestLoggerMethods.create_logger(None, True).close()
